"""Module that contains the code to run the core game engine.
You can create an instance directly by using the `ZombieDieGame`.

```python
from zombie_nomnom import ZombieDieGame, DrawDice, Score

score = Score()
draw_dice = DrawDice(3)

game = ZombieDieGame(players=["Mellow"])

game.process_command(draw_dice)
game.process_command(score)

```

For the core of the engine it doesn't actually know how to play but only 
how to manage a turn based on tracking the changes in a round with a `RoundState` object.

This allows you to extend the game by creating your own custom actions with the `Command` class.
This class defines the `execute` method and then allows you to specify what that action does to the game state.
You will need to return an new instance of `RoundState` objects that represents the effect of game state after the round is over.

```python
from zombie_nomnom import Command, ZombieDieGame

class CustomCommand(Command):
    def execute(self, round: RoundState) -> RoundState:
        # do meaningful work to define what you want this command to do to the round.
        return round  # return either a new command or the exact same command unchanged.

custom_command = CustomCommand()
game = ZombieDieGame(players=["Meelow"])

game.process_command(custom_command) # now it just works in the game!!

```

This only allows you to modify the state of the current player or the current turn that is active in the engine.
That being said it should provide a nice way to extend the app with custom actions for a players.

The core objects that we use in our engine are three:
- `Player`
- `RoundState`

```python
from zombie_nomnom.models import DieBag, Die, Face
from zombie_nomnom.engine import Player, RoundState

player = Player(
    name="Mega Man"
)

# you can add dice to their hand.
player_with_die_added = player.add_dice(
    Die(
        faces=[Face.BRAIN] * 6
    )
)
player_with_die_added.hand # dice is now updated with new die.

player.rerolls # pulls the dice that are re-rollable in hand
player.brains # pulls all the scoring dice in hand.
player.shots # pulls all the damaging dice in hand.

# counts the shots in the hand and see if you have been shot 3 or more times.
player.is_player_dead() # player is not dead

# new player with nothing in their hand.
new_player = player.clear_hand()
new_player.hand # is empty list

# resets the players game stats
new_player = player.reset() 
new_player.total_brains # now 0
new_player.hand # empty list

# counts the dice in the hand and adds to score.
new_player = player.calculate_score()
new_player.total_brains # player had 1 brain so it will now be 1
new_player.hand # is not empty list

# no methods just holds value to package it in a single object.
round = RoundState(
    bag=DieBag.standard_bag(),
    player=player,
)
```
"""

from .models import Player, RoundState, uuid_str
from .commands import Command, Score, DrawDice
from .game import ZombieDieGame


__all__ = [
    "Player",
    "RoundState",
    "uuid_str",
    "Command",
    "Score",
    "DrawDice",
    "ZombieDieGame",
    "commands",
    "game",
    "models",
    "serialization",
]



import pytest
from zombie_nomnom.models.bag import DieBag
from zombie_nomnom.models.dice import Die, Face, DieFace
from zombie_nomnom.engine import (
    DrawDice,
    Player,
    RoundState,
    Score,
    ZombieDieGame,
    Command,
    uuid_str,
)

from pydantic import ValidationError


def test_uuid_str__when_generating_new_id__returns_str():
    value = uuid_str()

    assert isinstance(value, str), f"Value we got back was not a str: {value}"


def test_player_score__when_reset__resets_game_info_but_keeps_identity():
    # act
    sut = Player(
        name="Bill The Kid",
        hand=[create_die()],
        total_brains=69,
    )

    new_player = sut.reset()

    # assert
    assert new_player.name == sut.name
    assert new_player.id == sut.id
    assert new_player.total_brains == 0
    assert new_player.hand == []


def test_player_score__when_clearing_hand__returns_hand_empty():
    sut = Player(
        name="Sergie",
        total_brains=22,
        hand=[
            create_die(),
        ],
    )

    new_player = sut.clear_hand()
    assert new_player is not sut, "Returned the same player."
    assert new_player.total_brains == sut.total_brains, "Score was lost in translation."
    assert new_player.hand == [], "Hand was not cleared."
    assert new_player.id == sut.id, "Created a new player id."
    assert new_player.name == sut.name, "Changed name."


def test_player_score__when_adding_die__adds_to_hand():
    # arrange
    existing_player = Player(name="tester", total_brains=0, hand=[])
    die = Die(faces=[Face.BRAIN] * 6)
    # act
    sut = existing_player.add_dice(die)

    # assert

    assert sut.hand == [die]


def test_player_score__when_adding_dice__keeps_dice_in_hand():
    # arrange
    sut = Player(
        name="tester",
        total_brains=0,
        hand=[create_die(Face.BRAIN)] * 3,
    )

    # act
    sut = sut.add_dice(create_die(Face.FOOT))

    # assert

    assert len(sut.hand) == 4


def test_player_score__when_scoring__counts_all_brains():
    sut = Player(
        name="Medical",
        hand=[
            create_die(Face.BRAIN),
            create_die(Face.FOOT),
            create_die(Face.SHOTGUN),
        ],
        total_brains=0,
    )

    scored = sut.calculate_score()

    assert scored.total_brains == 1


def test_player_score__when_scoring_hand_with_no_brains__does_not_change_score():
    sut = Player(
        name="Medical",
        hand=[
            create_die(Face.FOOT),
            create_die(Face.FOOT),
            create_die(Face.SHOTGUN),
        ],
        total_brains=0,
    )

    scored = sut.calculate_score()

    assert scored.total_brains == 0


def test_player_score__when_scoring_hand_with_existing_points__adds_points_together():
    sut = Player(
        name="Medical",
        hand=[
            create_die(Face.BRAIN),
            create_die(Face.FOOT),
            create_die(Face.SHOTGUN),
        ],
        total_brains=11,
    )

    scored = sut.calculate_score()

    assert scored.total_brains == 12


def test_players_score__when_scoring_with_custom_face__adds_all_score_per_face():
    sut = Player(
        name="Medici",
        hand=[
            create_die(Face.BRAIN),
            create_die(DieFace(name="Super Brains", score=13, damage=0)),
        ],
        total_brains=0,
    )
    assert sut.calculate_score().total_brains == 14


def create_die(selected_face: Face | DieFace | None = None):
    return Die(
        faces=[selected_face or Face.SHOTGUN] * 6,
        current_face=selected_face,
    )


def test_player_score__when_hand_has_three_shotguns__player_death():
    sut = Player(
        name="death",
        hand=[
            create_die(Face.SHOTGUN),
            create_die(Face.SHOTGUN),
            create_die(Face.SHOTGUN),
        ],
    )

    assert sut.is_player_dead(), "He isn't dead bobby"


def test_player_score__when_hand_has_two_shotguns__player_is_alive():
    sut = Player(
        name="death",
        hand=[
            create_die(Face.SHOTGUN),
            create_die(Face.SHOTGUN),
        ],
    )

    assert not sut.is_player_dead(), "He isn't alive bobby"


def test_player_score__when_player_draws_instakill__player_is_dead():
    sut = Player(
        name="death",
        hand=[
            create_die(DieFace(name="instakill", score=0, damage=3)),
            create_die(Face.BRAIN),
            create_die(Face.BRAIN),
            create_die(Face.BRAIN),
        ],
    )
    assert sut.is_player_dead(), "He isn't dead bobby"


def test_zombie_die_game__when_creating_game_with_no_players__raises_exception():
    with pytest.raises(ValueError):
        ZombieDieGame(
            players=[],
        )


def test_zombie_die_game__when_resetting_players__sets_current_player_to_none():
    basic_game = ZombieDieGame(
        players=["Player One", "Player Two"],
    )

    # arrange
    basic_game.current_player = 1

    # act
    basic_game.reset_players()

    # assert
    assert basic_game.current_player is None


def test_zombie_die_game__when_resetting_players__each_player_is_reset():
    sut = ZombieDieGame(players=["David", "Ackerman"])

    sut.reset_players()

    assert all(player.total_brains == 0 and player.hand == [] for player in sut.players)


def test_zombie_die_game__when_resetting_game__calls_resets_all_game_state():
    sut = ZombieDieGame(players=["Jiren"])

    sut.reset_game()

    assert len(sut.players) == 1
    assert not sut.round.ended
    assert sut.current_player == 0


def test_zombie_die_game__when_round_transitioned__uses_new_round():
    known_player = Player(name="Gray Man")
    sut = ZombieDieGame(
        players=[known_player],
        round=RoundState(
            bag=DieBag.standard_bag(),
            player=known_player,
        ),
        current_player=0,
    )
    old_round = sut.round

    # next round
    sut.next_round()

    assert sut.round == old_round


def test_zombie_die_game__when_round_transitioned__transitions_player():
    sut = ZombieDieGame(players=["Player One", "Player Two"])

    sut.next_round()

    assert sut.current_player == 0


def test_zombie_die_game__when_round_transitioned__transitions_player():
    sut = ZombieDieGame(
        players=["Player One", "Player Two"],
        current_player=1,
    )

    sut.next_round()

    assert sut.current_player == 0


def test_zombie_die_game__when_transitioning_round__updates_the_current_player_to_next():
    sut = ZombieDieGame(players=["First", "Second"], current_player=0)

    sut.next_round()

    assert sut.current_player == 1


def test_zombie_die_game__when_transitioning_round_and_currently_on_last_player__next_player_is_first_player():
    sut = ZombieDieGame(
        players=["Milo", "Xander"],
        current_player=1,
    )

    sut.next_round()

    assert sut.current_player == 0


def test_zombie_die_game__when_transitioning_round_for_single_player__next_player_is_first_player():
    sut = ZombieDieGame(
        players=["Dean"],
        current_player=0,
    )

    sut.next_round()

    assert sut.current_player == 0


class NoOpCommand(Command):
    def execute(self, round: RoundState) -> RoundState:
        return round


def test__zombie_die_game__process_command__raises_value_eror_when_game_over_is_true():
    sut = ZombieDieGame(players=["Lily"], game_over=True)

    with pytest.raises(ValueError):
        sut.process_command(NoOpCommand())


def test_zombie_die_game__when_processing_command__stores_command_in_commands():
    sut = ZombieDieGame(players=["Game"])

    assert not sut.commands

    sut.process_command(NoOpCommand())

    assert sut.commands


def test_zombie_die_game__when_determining_game_over__does_not_end_game_when_player_is_ahead_of_winning_player():
    turn_player = Player(
        name="Two",
        total_brains=11,
    )
    sut = ZombieDieGame(
        players=[
            Player(name="Player", total_brains=15),
            turn_player,
            Player(name="Tres", total_brains=12),
        ],
        round=RoundState(
            bag=DieBag.standard_bag(),
            player=turn_player,
            ended=True,
        ),
        current_player=1,
        first_winning_player=0,
    )

    sut.check_for_game_over()

    assert not sut.game_over


def test_zombie_die_game__when_determining_game_over_on_an_active_round__does_not_end_game():
    turn_player = Player(
        name="Two",
        total_brains=11,
    )
    sut = ZombieDieGame(
        players=[
            Player(name="Player", total_brains=15),
            turn_player,
            Player(name="Tres", total_brains=12),
        ],
        round=RoundState(
            bag=DieBag.standard_bag(),
            player=turn_player,
            ended=False,
        ),
        current_player=1,
        first_winning_player=0,
    )

    sut.check_for_game_over()

    assert not sut.game_over


def test_zombie_die_game__when_determining_game_over_and_first_player_won_and_last_players_turn_just_ended__ends_game():
    turn_player = Player(
        name="Two",
        total_brains=11,
    )
    sut = ZombieDieGame(
        players=[
            Player(name="Player", total_brains=15),
            turn_player,
            Player(name="Tres", total_brains=12),
        ],
        round=RoundState(
            bag=DieBag.standard_bag(),
            player=turn_player,
            ended=True,
        ),
        current_player=2,
        first_winning_player=0,
    )

    sut.check_for_game_over()

    assert sut.game_over


def test_zombie_die_game__when_determining_game_over_and_first_player_won_and_last_players_turn_still_going__does_not_end_game():
    turn_player = Player(
        name="Two",
        total_brains=11,
    )
    sut = ZombieDieGame(
        players=[
            Player(name="Player", total_brains=15),
            turn_player,
            Player(name="Tres", total_brains=12),
        ],
        round=RoundState(
            bag=DieBag.standard_bag(),
            player=turn_player,
            ended=False,
        ),
        current_player=2,
        first_winning_player=0,
    )

    sut.check_for_game_over()

    assert not sut.game_over


def test_zombie_die_game__when_determining_game_over_and_next_turn_player_was_the_first_winner_and_current_turn_over__ends_game():
    turn_player = Player(
        name="Two",
        total_brains=11,
    )
    sut = ZombieDieGame(
        players=[
            Player(name="Player", total_brains=15),
            turn_player,
            Player(name="Tres", total_brains=12),
        ],
        round=RoundState(
            bag=DieBag.standard_bag(),
            player=turn_player,
            ended=True,
        ),
        current_player=1,
        first_winning_player=2,
    )

    sut.check_for_game_over()

    assert sut.game_over


def test_zombie_die_game__when_determining_game_over_for_single_player_game__ends_game():
    turn_player = Player(
        name="Two",
        total_brains=15,
    )
    sut = ZombieDieGame(
        players=[
            turn_player,
        ],
        round=RoundState(
            bag=DieBag.standard_bag(),
            player=turn_player,
            ended=True,
        ),
        current_player=0,
        first_winning_player=None,
    )

    sut.check_for_game_over()

    assert sut.game_over


def test_zombie_die_game__when_determining_game_over_for_two_player_game__ends_game():
    turn_player = Player(
        name="Two",
        total_brains=15,
    )
    sut = ZombieDieGame(
        players=[
            turn_player,
            Player(name="Uno", total_brains=14),
        ],
        round=RoundState(
            bag=DieBag.standard_bag(),
            player=turn_player,
            ended=True,
        ),
        current_player=0,
        first_winning_player=1,
    )

    sut.check_for_game_over()

    assert sut.game_over


def test_zombie_die_game__when_determining_game_over_for_two_player_game_last_player_finishes__ends_game():
    turn_player = Player(
        name="Two",
        total_brains=15,
    )
    sut = ZombieDieGame(
        players=[
            Player(name="Uno", total_brains=14),
            turn_player,
        ],
        round=RoundState(
            bag=DieBag.standard_bag(),
            player=turn_player,
            ended=True,
        ),
        current_player=1,
        first_winning_player=0,
    )

    sut.check_for_game_over()

    assert sut.game_over


def test_zombie_die_game__when_resolving_winner__selects_player_with_highest_brain_count():
    turn_player = Player(
        name="Two",
        total_brains=15,
    )
    sut = ZombieDieGame(
        players=[
            Player(name="Uno", total_brains=14),
            turn_player,
            Player(
                name="Winner",
                total_brains=22,
            ),
        ],
        round=RoundState(
            bag=DieBag.standard_bag(),
            player=turn_player,
            ended=True,
        ),
        current_player=1,
        first_winning_player=0,
    )

    assert sut.winner.name == "Winner"


def all_face_bag(
    face: Face | None = None,
    amount_in_bag: int = 12,
):
    return DieBag(
        dice=[create_die(face)] * amount_in_bag,
    )


def test_draw_dice__when_given_a_valid_round__draws_dice_and_rolls_them():
    sut = DrawDice()
    player = Player(name="Ready Player One", hand=[], total_brains=0)
    round_info = RoundState(
        bag=all_face_bag(Face.BRAIN),
        player=player,
        ended=False,
    )

    new_info = sut.execute(round_info)

    assert isinstance(
        new_info, RoundState
    ), f"Was not give a round state but instead {type(new_info)}"
    new_player = new_info.player
    assert new_player.name == player.name
    assert new_player.hand, "Hand is empty"


def test_draw_dice__when_drawing_dice__only_gets_three_from_bag():
    sut = DrawDice()
    player = Player(name="Ready Player One", hand=[], total_brains=0)
    round = RoundState(
        bag=all_face_bag(Face.BRAIN),
        player=player,
        ended=False,
    )

    new_round = sut.execute(round)
    old_bag = round.bag
    new_bag = new_round.bag
    assert (
        len(old_bag) - len(new_bag) == 3
    ), f"you have pull not 3 dice: {len(old_bag) - len(new_bag)}"


def test_draw_dice__when_drawing_dice__check_if_player_is_dead():
    sut = DrawDice()
    player = Player(name="Ready Player One", hand=[], total_brains=0)
    round_info = RoundState(
        bag=all_face_bag(Face.BRAIN),
        player=player,
        ended=False,
    )

    new_info = sut.execute(round_info)
    new_player = new_info.player
    assert new_player.is_player_dead() == new_info.ended


def test_draw_dice__when_round_is_already_over__returns_round_as_is():
    sut = DrawDice()
    bag = all_face_bag(Face.BRAIN).draw_dice()
    player = Player(name="Ready Player One", hand=bag.drawn_dice, total_brains=0)
    round_info = RoundState(
        bag=bag,
        player=player,
        ended=True,
    )

    new_info = sut.execute(round_info)

    assert new_info is round_info, "Created a new round when it should not have."


def test_draw_dice__when_give_not_a_goddamn_round__raises_validation_error():
    sut = DrawDice()
    with pytest.raises(ValidationError):
        sut.execute(object())


def test_draw_dice__when_creating_command_with_negative_value__raises_exception():
    with pytest.raises(ValueError):
        DrawDice(amount_drawn=-1)


def test_draw_dice__when_creating_command_with_zero__raises_exception():
    with pytest.raises(ValueError):
        DrawDice(amount_drawn=0)


def test_draw_dice__when_rolled_death_for_player__round_is_ended_and_players_hand_is_cleared():
    sut = DrawDice()
    # Damn he is about to throw so hard...
    round = RoundState(
        bag=all_face_bag(Face.SHOTGUN, amount_in_bag=4),
        player=Player(
            name="Billy",
            hand=[
                create_die(Face.BRAIN),
                create_die(Face.BRAIN),
                create_die(Face.BRAIN),
            ],
        ),
        ended=False,
    )

    new_round = sut.execute(round)

    assert new_round.player.hand == []
    assert new_round.ended is True


def test_draw_dice__when_rolling_dice__uses_dice_in_hand_first():
    round = RoundState(
        bag=all_face_bag(Face.BRAIN, amount_in_bag=10),
        player=Player(
            name="OuiOui Bagguette",
            hand=[
                create_die(Face.FOOT),
                create_die(Face.FOOT),
                create_die(Face.FOOT),
            ],
        ),
        ended=False,
    )
    sut = DrawDice()

    new_round = sut.execute(round)

    assert new_round.bag.drawn_dice == []
    assert len(new_round.bag) == 10


def test_draw_dice__when_rolling_dice_with_feet_in_hand__fills_in_remaining_with_die_from_bag():
    round = RoundState(
        bag=all_face_bag(Face.BRAIN, amount_in_bag=10),
        player=Player(
            name="OuiOui Bagguette",
            hand=[
                create_die(Face.FOOT),
            ],
        ),
        ended=False,
    )
    sut = DrawDice()

    new_round = sut.execute(round)

    assert len(new_round.bag.drawn_dice) == 2
    assert len(new_round.bag) == 8


def test_draw_dice__when_trying_to_roll_die_with_not_enough_dice_in_bag__copies_brains_into_bag_and_rolls_them():
    round = RoundState(
        bag=all_face_bag(Face.BRAIN, amount_in_bag=1),
        player=Player(
            name="OuiOui Bagguette",
            hand=[
                create_die(Face.FOOT),
                create_die(Face.BRAIN),
                create_die(Face.BRAIN),
            ],
        ),
        ended=False,
    )
    sut = DrawDice()

    new_round = sut.execute(round)

    assert len(new_round.bag.drawn_dice) == 2
    assert len(new_round.bag) == 1


def test_score__when_scoring__calculates_based_on_players_hand():
    sut = Score()
    player = Player(name="Billy the Goat")
    round = RoundState(
        bag=DieBag.standard_bag(),
        player=player,
        ended=False,
    )
    new_round = sut.execute(round)

    assert new_round.player.total_brains == 0, "Did not calculate score correctly."


def test_score__when_scoring_already_ended_round__returns_same_round():
    sut = Score()
    player = Player(name="Billy the Goat")
    round = RoundState(bag=DieBag.standard_bag(), player=player, ended=True)

    new_round = sut.execute(round)

    assert new_round is round



import uuid

from pydantic import BaseModel, Field

from zombie_nomnom.models.bag import DieBag
from zombie_nomnom.models.dice import Die, Face, DieFace


def uuid_str() -> str:
    """Creates a stringified uuid that we can use.

    **Returns**
    - `str`: stringified uuid
    """
    return str(uuid.uuid4())


def is_scoring_face(face: Face | DieFace) -> bool:
    """Checks if a face is scoring.

    A scoring face is a face that will score points for the player.
    This is useful for determining if a player should continue rolling or not.

    **Arguments**
    - `face: Face | DieFace`: The face to check.

    **Returns**
    - `bool`: If the face is a scoring face.
    """
    return (
        isinstance(face, Face)
        and face == Face.BRAIN
        or isinstance(face, DieFace)
        and face.score
    )


def is_damaging_face(face: Face | DieFace) -> bool:
    """Checks if a face is damaging.

    A damaging face is a face that will limit the amount of dice the player can roll.
    This is useful for determining if a player should stop rolling or not.

    **Arguments**
    - `face: Face | DieFace`: The face to check.

    **Returns**
    - `bool`: If the face is a damaging face.
    """
    return (
        isinstance(face, Face)
        and face == Face.SHOTGUN
        or isinstance(face, DieFace)
        and face.damage
    )


def is_blank_face(face: Face | DieFace) -> bool:
    """Checks if a face is blank.

    A blank face is a face that doesn't score any points nor does it damage the player.
    This is useful for determining if a die should be rolled again.

    **Arguments**
    - `face: Face | DieFace`: The face to check.

    **Returns**
    - `bool`: True if the face is blank.
    """
    return (
        isinstance(face, Face)
        and face == Face.FOOT
        or isinstance(face, DieFace)
        and not face.score
        and not face.damage
    )


class Player(BaseModel):
    """Player in the game. This manages the total points
    they have and the dice they pulled from the bag. Has
    methods to manage the hand and then calculate points.
    Also has ways to be able to tell if players turn should finish.
    """

    id: str = Field(default_factory=uuid_str)
    """id field used to identify the player more unqiuely"""
    name: str
    """name of the player that we display on the screen"""
    total_brains: int = 0
    """total points they currently have in the game"""
    hand: list[Die] = []
    """the dice the player is currently holding"""

    @property
    def rerolls(self) -> list[Die]:
        """A view of the hand that shows all the dice that are re-rollable.

        **Returns**:
        - `list[zombie_nomnom.Die]`: re-rollable dice
        """
        return [die for die in self.hand if is_blank_face(die.current_face)]

    @property
    def brains(self) -> list[Die]:
        """A view of the hand that shows all the dice that scores points.

        **Returns**
        - `list[zombie_nomnom.Die]`: scoreable dice
        """
        return [die for die in self.hand if is_scoring_face(die.current_face)]

    @property
    def shots(self) -> list[Die]:
        """A view of the hand that shows the shots you have taken.

        **Returns**
        - `list[zombie_nomnom.Die]`: damaging dice
        """
        return [die for die in self.hand if is_damaging_face(die.current_face)]

    def is_player_dead(self) -> bool:
        """Calculates whether or not the player should be dead based on the shots in their hand.

        **Returns**
        - bool: True when player is considered dead.
        """
        total = 0
        for shot in self.shots:
            face = shot.current_face
            if isinstance(face, DieFace):
                total += face.damage
            else:
                total += 1
        # if you have 3 shots you are dead XP
        return total >= 3

    def add_dice(self, *dice: Die) -> "Player":
        """Creates a new player adding the dice the caller gives plus any dice in their hand currently.

        **Returns**
        - `Player`: New player instance with updated hand.
        """
        return Player(
            id=self.id,
            name=self.name,
            hand=[*self.hand, *dice],
            total_brains=self.total_brains,
        )

    def clear_hand(self) -> "Player":
        """Creates a new player with no dice in their hand.

        **Returns**
        - `Player`: New player instance with empty hand.
        """
        return Player(
            id=self.id,
            name=self.name,
            total_brains=self.total_brains,
        )

    def reset(self) -> "Player":
        """Creates a new player instance with all the game related fields set back to their default values.

        **Returns**
        - `Player`: New player instance with all values reset.
        """
        return Player(
            id=self.id,
            name=self.name,
            total_brains=0,
            hand=[],
        )

    def calculate_score(self) -> "Player":
        """Creates a new player that has added the score of the current hand and leaves an empty hand.

        **Returns**
        - `Player`: New player instance with score adjusted for scoring dice in hand and hand reset back to empty list.
        """
        additional_score = 0
        for brain_xo in self.brains:
            face = brain_xo.current_face
            if isinstance(face, DieFace):
                additional_score += face.score
            else:
                additional_score += 1
        return Player(
            id=self.id,
            name=self.name,
            total_brains=additional_score + self.total_brains,
        )


class RoundState(BaseModel):
    """
    Object representing the state of a round in the game. Keeps track of the bag, player,
    and whether or not the round has ended.
    """

    bag: DieBag
    """Bag that is currently being played in the round"""
    player: Player
    """Player that is currently playing"""
    ended: bool = False
    """Records whether or not the current round is over"""


class DieRecipe(BaseModel):
    """The recipe for a dice to be used in the bag. Will have the array of faces
    that we will use for our die and how many instances to add to our bag.
    """

    faces: list[Face]
    """The faces that will make up the dice."""
    amount: int
    """The amount of dice that will be added to the bag."""



from copy import deepcopy
import operator
from typing import Callable


from zombie_nomnom.models import DieBag
from zombie_nomnom.models.dice import Die

from .models import DieRecipe, Player, RoundState
from .commands import Command


def bag_from_recipes(dice_recipes: list[DieRecipe]):
    """Function that can be used to create a bag_function from a list of dice recipes.

    **Parameters**
    - dice_recipes (`list[DieRecipe]`): the list of recipes to create the dice in the bag.

    **Returns**
    - `Callable[[], DieBag]`: The new closure that creates bags by referencing the dice_recipes array.

    **Raises**
    - `ValueError`: When give no recipes. Must have recipes to be able to create a bag that is non-empty.
    """
    if not dice_recipes:
        raise ValueError("Need recipes to build bag from.")

    def _bag_function():
        dice = []
        for recipe in dice_recipes:
            dice.extend(
                Die(
                    faces=deepcopy(recipe.faces),
                )
                for _ in range(recipe.amount)
            )
        return DieBag(
            dice=dice,
        )

    return _bag_function


class ZombieDieGame:
    """Instance of the zombie dice that that manages a bag of dice that will be used to coordinate how the game is played.

    **Parameters**
    - players (`list[PlayerScore]`): players in the game
    - commands (`list[tuple[Command, RoundState]]`, optional): previous commands that have been run before in the game. Defaults to `[]`
    - bag_function (`Callable[[], DieBag]`, optional): function that will generate a `zombie_nomnom.DieBag` that will be used in the round transitions. Defaults to `zombie_nomnom.DieBag.standard_bag`
    - score_threshold (`int`, optional): the score threshold that will trigger the end game. Defaults to `13`
    - current_player (`int | None` optional): the index in the player array to represent the current player. Defaults to `None`
    - first_winning_player (`int | None` optional): the index in the player array that represents the first player to meet or exceed the score threshold. Defaults to `None`
    - game_over (`bool`, optional): marks whether or not game is over. Defaults to `False`
    - round (`RoundState | None`, optional): the current round of the game being played. Defaults to a new instance that is created for the first player in the player array.
    - bag_recipes (`list[DieRecipe]`):

    **Raises**
    - `ValueError`: When there is not enough players to play a game.
    """

    players: list[Player]
    """Players that are in the game."""
    commands: list[tuple[Command, RoundState]]
    """Commands that have been processed in the game and the round state they were in before it started."""
    bag_function: Callable[[], DieBag]
    """Function that we use when we need to create a new bag for a round."""
    round: RoundState | None
    """Current round that we are on."""
    current_player: int | None
    """Index of player in the players array who's turn it currently is."""
    first_winning_player: int | None
    """Index of player who first exceeded or matched the `score_threshold`."""
    game_over: bool
    """Marker for when the game is over."""
    score_threshold: int
    """Threshold required for a player to start the end game."""
    bag_recipes: list[DieRecipe]

    def __init__(
        self,
        players: list[str | Player],
        commands: list[tuple[Command, RoundState]] | None = None,
        bag_function: Callable[[], DieBag] | None = None,
        score_threshold: int = 13,
        current_player: int | None = None,
        first_winning_player: int | None = None,
        game_over: bool = False,
        round: RoundState | None = None,
        bag_recipes: list[DieRecipe] | None = None,
    ) -> None:
        if len(players) == 0:
            raise ValueError("Not enough players for the game we need at least one.")

        self.commands = list(commands) if commands else []
        self.players = [
            (
                Player(name=name_or_score)
                if isinstance(name_or_score, str)
                else name_or_score
            )
            for name_or_score in players
        ]
        if not bag_recipes:
            self.bag_function = bag_function or DieBag.standard_bag
            self.bag_recipes = []
        else:
            self.bag_function = bag_from_recipes(bag_recipes)
            self.bag_recipes = bag_recipes
        self.score_threshold = score_threshold

        self.round = round
        self.current_player = current_player
        self.first_winning_player = first_winning_player
        self.game_over = game_over

        if self.round is None and self.current_player is None:
            self.next_round()

    @property
    def winner(self) -> Player:
        """The player with the highest score in the players array. On Ties uses the player with the lowest index.

        **Returns**
        - `Player`: The winning player
        """
        return max(self.players, key=operator.attrgetter("total_brains"))

    def reset_players(self):
        """Resets the game state so that players scores are set to zero and the current_player is reset to `None` as well as the first_winning_player"""
        self.players = [player.reset() for player in self.players]
        self.current_player = None
        self.first_winning_player = None

    def reset_game(self):
        """Resets the game state by resetting the players, clearing commands, and transitioning to the next round which will be the first players round."""
        self.reset_players()
        self.commands = []
        self.next_round()

    def next_round(self):
        """Transitions the current player index point to the next players turn and then sets the round field with that player and a new bag to play the round."""
        if self.current_player is not None and self.round:
            self.players[self.current_player] = self.round.player

        if self.current_player is None:
            self.current_player = 0
        elif self.current_player + 1 < len(self.players):
            self.current_player = self.current_player + 1
        else:
            self.current_player = 0
        self.round = RoundState(
            bag=self.bag_function(),
            player=self.players[self.current_player],
            ended=False,
        )

    def check_for_game_over(self) -> bool:
        """Checks if game is over and sets the game_over field."""
        if not self.round.ended:
            return  # Still not done with their turns.
        game_over = False
        # GAME IS OVER WHEN THE LAST PLAYER IN A ROUND TAKES THERE TURN
        # I.E. IF SOMEONE MEETS THRESHOLD AND LAST PLAYER HAS HAD A TURN
        if len(self.players) == 1 and self.winner.total_brains >= self.score_threshold:
            game_over = True

        if self.first_winning_player is None:
            if self.players[self.current_player].total_brains >= self.score_threshold:
                self.first_winning_player = self.current_player
        else:
            if (
                self.first_winning_player == 0
                and self.current_player == len(self.players) - 1
            ):
                game_over = True
            elif (
                self.first_winning_player > self.current_player
                and self.first_winning_player - self.current_player == 1
            ):
                game_over = True

        self.game_over = game_over

    def update_player(self):
        """Updates the player in the players array with the instance that is currently on the round field for the current player index."""
        self.players[self.current_player] = self.round.player

    def process_command(self, command: Command) -> RoundState:
        """Applies the given command to the active round and transitions to the next round if the current round is over.

        **Parameters**
        - command (`Command`): command that will modify the round state.

        **Raises**
        - `ValueError`: When trying to process a command when the game is already over.

        **Returns**
        - `RoundState`: The round information that happened due to the command.
        """
        if self.game_over:
            raise ValueError("Cannot command an ended game please reset game.")

        self.commands.append((command, self.round))

        resulting_round = command.execute(self.round)
        self.round = resulting_round
        if self.round.ended:
            self.update_player()
            self.check_for_game_over()
            self.next_round()
        return resulting_round



from unittest.mock import patch

import pytest


@pytest.fixture
def patch_random_randint():
    with patch("random.randint") as mock_random:
        yield mock_random


@pytest.fixture
def patch_random_choice():
    with patch("random.choice") as mock_choice:
        yield mock_choice



from unittest.mock import Mock

import pytest
from zombie_nomnom.cli import StrippedStr


@pytest.fixture
def mock_param():
    return Mock(name="param")


@pytest.fixture
def mock_ctx():
    return Mock(name="context")


class StaticStrObject:
    def __str__(self) -> str:
        return "Static"


class ObjectWithSpacesInStr:
    def __str__(self) -> str:
        return " spaces "


@pytest.mark.parametrize(
    "value,expected_value",
    [
        (" string with leading spaces", "string with leading spaces"),
        ("string with trailing spaces ", "string with trailing spaces"),
        (" string with both ", "string with both"),
        (StaticStrObject(), "Static"),
        (ObjectWithSpacesInStr(), "spaces"),
    ],
)
def test_stripped_str__when_given_value__formats_string_as_expected(
    value, expected_value, mock_ctx, mock_param
):
    sut = StrippedStr()

    actual_value = sut.convert(value, mock_param, mock_ctx)

    assert actual_value == expected_value



"""
Module that contains the click entrypoint for our cli interface.

Currently this only contains code to run the game form the cli but this will be extended 
to also run this as a web app including a built in server with react spa app.
"""

import click
from .cli import run_game


@click.group()
def main():
    """
    main group that represents the top-level: ***zombie-nomnom***

    This will be used to decorate sub-commands for zombie-nomnom.

    ***Example Usage:***
    ```python
    @main.command("sub-command")
    def sub_command():
        # do actual meaningful work.
        pass
    ```
    """
    pass


@main.command("cli")
def cli():
    """
    Command to start the zombie_dice game from the command line.
    """
    run_game()

    # ask after we finish a single game assume they will quit when they want to.
    while click.confirm(text="Play another game of zombie dice?"):
        run_game()

    click.echo("Thank you for playing!!")



import pytest

from zombie_nomnom.engine.models import Face
from zombie_nomnom.engine.commands import DrawDice, Score
from zombie_nomnom.engine.serialization import (
    format_command,
    parse_command_dict,
    format_to_json_dict,
    parse_game_json_dict,
    ZombieDieGame,
    DieRecipe,
)


@pytest.fixture
def game():
    zombie_game = ZombieDieGame(players=["Player Uno"])
    zombie_game.bag_recipes = [DieRecipe(faces=[Face.FOOT] * 6, amount=3)]
    zombie_game.bag_function = DrawDice(amount_drawn=3)
    yield zombie_game


def test_format_command__when_formatting_known_command__uses_fully_qualified_name_for_class():
    command = Score()
    expected = {
        "cls": "zombie_nomnom.engine.commands.Score",
        "args": [],
        "kwargs": {},
    }
    actual = format_command(command)

    assert actual == expected


def test_format_command__when_formatting_command_with_args__puts_args_as_kwargs():
    command = DrawDice(amount_drawn=3)
    expected = {
        "cls": "zombie_nomnom.engine.commands.DrawDice",
        "args": [],
        "kwargs": {
            "amount_drawn": 3,
        },
    }
    actual = format_command(command)

    assert actual == expected


def test_parse_command_dict__when_parsing_will_load_class_and_initialize_correctly__loads_command_class():
    cmd_dict = {
        "cls": "zombie_nomnom.engine.commands.Score",
        "args": [],
        "kwargs": {},
    }

    actual = parse_command_dict(cmd_dict)

    assert isinstance(actual, Score)


def test_parse_command_dict__when_parsing_command_with_parameters_kwargs__loads_command_class():
    cmd_dict = {
        "cls": "zombie_nomnom.engine.commands.DrawDice",
        "args": [],
        "kwargs": {"amount_drawn": 3},
    }

    actual = parse_command_dict(cmd_dict)

    assert isinstance(actual, DrawDice)
    assert actual.amount_drawn == 3


def test_parse_command_dict__when_parsing_command_with_parameters_args__loads_command_class():
    cmd_dict = {
        "cls": "zombie_nomnom.engine.commands.DrawDice",
        "args": [3],
        "kwargs": {},
    }

    actual = parse_command_dict(cmd_dict)

    assert isinstance(actual, DrawDice)
    assert actual.amount_drawn == 3


def test_format_to_json_dict__when_bag_function_is_none_and_bag_recipes_is_empty__assumes_standard_bag():
    game = ZombieDieGame(players=["Player Uno"])
    json = format_to_json_dict(game)
    assert json["bag_function"] == "standard"


def test_format_to_json_dict__when_bag_function_is_not_none_and_bag_recipes_exists__returns_valid_dict(
    game,
):
    sut = format_to_json_dict(game)
    assert sut
    assert isinstance(sut, dict)
    assert sut["bag_function"]
    assert sut["players"]


def test_parse_game_json_dict__when_given_valid_dict__returns_game_instance(
    game,
):
    game_dict = format_to_json_dict(game)
    sut = parse_game_json_dict(game_dict)

    assert sut
    assert sut.bag_recipes == game.bag_recipes
    assert sut.players == game.players
    assert sut.score_threshold == game.score_threshold
    assert sut.round == game.round


def test_parse_game_json_dict__when_bag_function_is_str_and_bag_function_is_not_standard_bag__raises_value_error(
    game,
):
    with pytest.raises(ValueError):
        game_dict = format_to_json_dict(game)
        game_dict["bag_function"] = "DrawDice"

        parse_game_json_dict(game_dict)



"""
.. include:: ../README.md
   :start-line: 2
   :end-before: Contribution
"""

from .models.bag import DieBag
from .models.dice import Die, DieColor, Face, create_die
from .engine import Command, ZombieDieGame, Player, RoundState

__version__ = "1.0.0"


__all__ = [
    "app",
    "models",
    "engine",
    "cli",
    "DieBag",
    "Die",
    "DieColor",
    "Face",
    "create_die",
    "ZombieDieGame",
    "Player",
    "RoundState",
    "Command",
]



from abc import ABC, abstractmethod
import sys

from pydantic import validate_call
from .models import RoundState


class Command(ABC):
    """
    Used to modify round state. Cannot be used to reset game.
    """

    @abstractmethod
    def execute(self, round: RoundState) -> RoundState:  # pragma: no cover
        """
        Method to generate a new RoundState that represents modifications on the command.

        **Parameters**
        - round(`RoundState`): the round we are on.

        **Returns** `RoundState`

        New instance of round with modified state.
        """


class DrawDice(Command):
    """
    The core command that represents handling a draw action in the game.
    This will attempt to draw dice in your hand to fill in any dice that
    are not re-rollable then roll the dice for the turn. Then it will
    check to make sure you are still alive and if so keep dice in your hand
    and return a new round object.

    **Parameters**
    - amount_drawn (`int`): Dice this action will attempt to draw.
    """

    amount_drawn: int
    """Amount of dice that this action will attempt to draw."""

    def __init__(self, amount_drawn: int = 3) -> None:
        if amount_drawn <= 0:
            raise ValueError("Cannot draw a no or a negative amount of dice.")
        self.amount_drawn = amount_drawn

    @validate_call
    def execute(self, round: RoundState) -> RoundState:
        """
        Executes a dice draw on a round that is active.

        If round is already over will return given round context.

        **Parameters**
        - round(`RoundState`): the round we are on.

        **Returns** `RoundState`

        New instance of a round with player adding dice to hand.
        """
        if round.ended:
            return round
        player = round.player
        dice_to_roll = player.rerolls
        total_dice = len(dice_to_roll)
        try:
            bag = (
                round.bag.clear_drawn_dice()
                if total_dice == self.amount_drawn
                else round.bag.draw_dice(amount=self.amount_drawn - total_dice)
            )
        except ValueError as exc:
            return self.execute(
                round=RoundState(
                    bag=round.bag.add_dice(player.brains),
                    player=player,
                    ended=round.ended,
                )
            )
        dice_to_roll.extend(bag.drawn_dice)
        player = player.add_dice(*bag.drawn_dice)

        for die in dice_to_roll:
            die.roll()

        ended = player.is_player_dead()
        if ended:
            player = player.clear_hand()

        return RoundState(
            bag=bag,
            player=player,
            ended=ended,
        )


class Score(Command):
    """
    Command to score the hand of a player and add the brains they have as points to their total.
    """

    def execute(self, round: RoundState) -> RoundState:
        """
        Scores the hand of the current player by rolling up all the scoring faces and adding it to their hand.

        **Parameters**
        - round (`RoundState`): The round we are currently in.

        **Returns** `RoundState`

        Roundstate that is now ended with the player with hand cleared and new score added to them.
        """
        if round.ended:
            return round
        player = round.player.calculate_score()
        return RoundState(
            bag=round.bag,
            player=player,
            ended=True,
        )



import pytest
from zombie_nomnom.engine.game import ZombieDieGame, bag_from_recipes
from zombie_nomnom.engine.models import DieRecipe
from zombie_nomnom.models.bag import DieBag
from zombie_nomnom.models.dice import Face


def test_bag_from_recipes__when_given_no_recipes__raises_value_error():
    with pytest.raises(ValueError):
        bag_from_recipes([])


def test_bag_from_recipes__when_given_single_recipe__returns_bag_that_only_creates_single_type_of_dice():
    recipes = [DieRecipe(amount=2, faces=[Face.BRAIN] * 6)]
    only_score_recipe = recipes[0]

    func = bag_from_recipes(recipes)
    new_bag = func()

    assert len(new_bag) == 2
    assert new_bag.dice[0].faces == only_score_recipe.faces


def test_zombie_die_game__when_creating_a_game_with_no_bag_function_but_recipes__creates_a_new_bag_func():
    sut = ZombieDieGame(
        players=["Skibbidy"],
        # the bag only contains rerolls mwuah-hahaha
        bag_recipes=[DieRecipe(faces=[Face.FOOT] * 6, amount=3)],
    )
    assert sut.bag_function is not None
    assert sut.bag_function is not DieBag.standard_bag
    assert sut.bag_recipes != []



"""
Module that contains the code related to the dice bag and how we draw and manage it.
This allows us to create custom sets per game of zombiedice by creating a method that creates `DieBag` objects.

For example of making your own diebag.
```python
from zombie_nomnom import DieBag, Die, Face

bag = DieBag(
    dice=[
        Die(faces=[Face.Brain] * 6), # Dice with only brains
        Die(faces=[Face.Brain] * 6),
        Die(faces=[Face.Brain] * 6),
        Die(faces=[Face.Brain] * 6),
    ]
)

# every bag action creates a new bag so you can store both 
new_bag = bag.draw_dice(1)

# drawing dice adds the die to your drawn_dice field
new_bag.drawn_dice

# You can check the amount of the dice in your bag using len()
len(new_bag) # 3 dice since this is the bag after you drew one.

# Clears the dice in the drawn_dice field and sets it to empty array
newer_bag = new_bag.clear_drawn_dice()

newer_bag.drawn_dice # is not empty array

# Bag also supports bool check which returns true if there are dice in the bag
bool(new_bag) # True since there are 3 dice

# We can also check if the bag is empty
new_bag.is_empty # False since there are 3 dice in the bag

# we can even add dice on the fly
bag.add_dice([Die(faces=[Face.Brain] * 6)])

len(bag) # will now be 5 since there are 5 dice.
```

If you don't want to bother making dice manually you can create our standard 13 dice bag:
```python
from zombie_nomnom import DieBag

# New bag containing: 6 Green Dice, 4 Yellow Dice, 3 Red Dice.
standard = DieBag.standard_bag()

len(standard) # will be 13 
```
"""

from copy import deepcopy
import random
from typing import Iterable

from pydantic import BaseModel

from .dice import Die, create_die, DieColor


class DieBag(BaseModel):
    """DieBag that contains and manages the dice in the game."""

    dice: list[Die]
    """The collection of dice in the bag"""
    drawn_dice: list[Die] = []
    """The dice the bag last drew"""

    @property
    def is_empty(self) -> bool:
        """The property that represents when the bag is considered empty.

        **Returns**
        - bool: True when bag is empty
        """
        return len(self) == 0

    def clear_drawn_dice(self) -> "DieBag":
        """Creates a new `DieBag` where its `drawn_dice` field is empty.

        **Returns**
        - `DieBag`: The new instance of DieBag without any drawn_dice.
        """
        return DieBag(
            dice=self.dice,
        )

    def add_dice(self, dice: Iterable[Die]) -> "DieBag":
        """Creates a new Diebag where you added the dice in this bag with
        the dice from the caller.

        **Parameters**
        - dice (`Iterable[zombie_nomnom.Die]`): Dice we are adding to the new bag.

        **Returns**
        - `DieBag`: The new bag of dice with both dice.
        """
        new_dice = [
            *(deepcopy(die) for die in self.dice),
            *(deepcopy(die) for die in dice),
        ]
        return DieBag(dice=new_dice, drawn_dice=[])

    def draw_dice(self, amount: int = 1) -> "DieBag":
        """Creates a new bag where we select dice from this one randomly
        and then set both the dice_drawn, and dice to a subset of the dice
        minus the dice we drew.

        **Parameters**
        - amount (`int`, optional): the total amount of dice we are getting from the bag. Defaults to 1.

        **Raises**
        - `ValueError`: Raised when there are not enough dice to be able to draw from the bag.

        **Returns**
            `DieBag`: The new bag with the dice_drawn set to the dice we drew from the bag.
        """
        if amount < 0 or amount > len(self):
            raise ValueError("The die bag does not have enough dice.")

        total = len(self)
        selected_dice = set()
        while len(selected_dice) < amount:
            selected_dice.add(random.randint(0, total - 1))
        return DieBag(
            dice=[
                die for index, die in enumerate(self.dice) if index not in selected_dice
            ],
            drawn_dice=[self.dice[index] for index in selected_dice],
        )

    def __len__(self):
        return len(self.dice)

    def __bool__(self):
        return len(self) > 0

    @classmethod
    def standard_bag(cls) -> "DieBag":
        """Creates a bag using the standard rules format.

        The bag will contain 6 GREEN dice, 4 YELLOW dice, and 3 RED dice.

        **Returns**
        - `DieBag`: the standard diebag that was created.
        """
        return cls(
            dice=[
                *(create_die(DieColor.GREEN) for _ in range(6)),
                *(create_die(DieColor.YELLOW) for _ in range(4)),
                *(create_die(DieColor.RED) for _ in range(3)),
            ],
        )



"""
Models
===

We contain the following models for the core of the game.

[`DieBag` information found here](models/bag#DieBag)
```python
from zombie_nomnom.models import DieBag, Die, DieColor, Face


# The core of the models is the bag which is our custom 
# collection that contains the code to select dice from the bag.
# It allows us to manage 
bag = DieBag(
    dice=[
        Die(
            faces=[
                Face.BRAIN,
                Face.BRAIN,
                Face.BRAIN,
                Face.SHOTGUN,
                Face.SHOTGUN,
                Face.SHOTGUN
            ]
        )
    ]
)
```

[`Die` information found here](models/dice#Die)
```python
from zombie_nomnom.models import Die


# Model to keep track of and roll the die we have 
# defined in the game.

custom_die = Die(
    faces=[
        Face.BRAIN,
        Face.BRAIN,
        Face.BRAIN,
        Face.SHOTGUN,
        Face.SHOTGUN,
        Face.SHOTGUN
    ]
)

```
"""

from .bag import DieBag
from .dice import Die, DieColor, Face

__all__ = [
    "DieBag",
    "Die",
    "DieColor",
    "Face",
    "bag",
    "dice",
]



import pytest
from zombie_nomnom.engine.commands import Score, DrawDice
from zombie_nomnom.engine.serialization import format_command, parse_command_dict


@pytest.mark.parametrize(
    "_cls,instance",
    [
        (Score, Score()),
        (DrawDice, DrawDice(3)),
    ],
)
def test__when_given_a_valid_command__can_format_then_parse_dict(
    _cls,
    instance,
):
    value_dict = format_command(instance)
    new_command = parse_command_dict(value_dict)

    assert isinstance(new_command, _cls)



"""This module contains the code that relates to how we can turn our game instances into a dictionary as well as
how we can store commands and load them back dynamically 
i.e. if you make custom commands how to load those commands back.

```python
from zombie_nomnom import ZombieDieGame
from zombie_nomnom.engine.serialization import format_to_json_dict

game = ZombieDieGame(players=["Player Uno"])

game_dict = format_to_json_dict(game)

# now you can write it as a yaml or json using whatever you like.
import json
with open('save_file.dat', 'w') as fp:
    json.dump(game_dict, fp) # write it out to the file!!

# at a later date...
from zombie_nomnom.engine.serialization import parse_game_json_dict
with open('save_file.dat', 'r') as fp:
    game_dict = json.load(fp)

# now we can keep playing like nothing happened!!
game = parse_game_json_dict(game_dict)
```

This allows you to control how you want to seralize the data and we store the format
in the TypedDict definitions within the module if you need to have some more information on how it looks.
This means that whatever way you wanna store these models you can and we can load it as long as the dict 
matches our defined structure.
"""

from enum import Enum
import importlib
from typing import Any, TypedDict

from zombie_nomnom.engine.commands import Command
from zombie_nomnom.engine.game import ZombieDieGame
from zombie_nomnom.engine.models import DieRecipe, Player, RoundState
from zombie_nomnom.models.bag import DieBag


class DieDict(TypedDict):
    sides: list[str]
    current_face: str | None


class PlayerDict(TypedDict):
    id: str
    name: str
    total_brains: int
    hand: list[DieDict]


class DieBagDict(TypedDict):
    dice: list[DieDict]
    drawn_dice: list[DieDict] | None


class RoundStateDict(TypedDict):
    diebag: DieBagDict


class CommandDict(TypedDict):
    cls: str
    args: list[Any]
    kwargs: dict[str, Any]


class DieRecipeDict(TypedDict):
    faces: list[str]
    amount: int


class ZombieDieGameDict(TypedDict):
    players: list[PlayerDict]
    commands: list[tuple[CommandDict, RoundStateDict]]
    current_player: int | None
    first_winning_player: int | None
    round: RoundStateDict
    game_over: bool
    score_threshold: int
    bag_function: str | list[DieRecipeDict]


# We may want this to be removed later idk yet?
class KnownFunctions(str, Enum):
    STANDARD = "standard"


def format_command(command: Command) -> CommandDict:
    cmd_type = type(command)
    module = cmd_type.__module__
    qual_name = cmd_type.__qualname__
    return {
        "cls": f"{module}.{qual_name}",
        "args": [],
        # only works if the field on the class matches the param in __init__.py
        "kwargs": command.__dict__,
    }


def parse_command_dict(command: CommandDict) -> Command:
    [*module_path, cls_name] = command.get("cls").split(".")
    module_name = ".".join(module_path)
    module = importlib.import_module(module_name)
    cls = getattr(module, cls_name)
    return cls(*command.get("args"), **command.get("kwargs"))


def format_to_json_dict(game: ZombieDieGame) -> ZombieDieGameDict:
    """Will default any game to be using the standard bag unless given a recipe to create dice.

    **Parameters**
    - game (`ZombieDieGame`): The game to serialize

    **Returns**
    - `ZombieDieGameDict`: The serialized game as a dict that is json serializable
    """
    return {
        "players": [player.model_dump(mode="json") for player in game.players],
        "bag_function": (
            KnownFunctions.STANDARD
            if not game.bag_recipes
            else [recipe.model_dump(mode="json") for recipe in game.bag_recipes]
        ),
        "commands": [
            (format_command(command), state.model_dump(mode="json"))
            for command, state in game.commands
        ],
        "current_player": game.current_player,
        "first_winning_player": game.first_winning_player,
        "game_over": game.game_over,
        "round": game.round.model_dump(mode="json"),
        "score_threshold": game.score_threshold,
    }


UNTRANSFORMED_KEYS = {
    "score_threshold",
    "current_player",
    "first_winning_player",
    "game_over",
}


def parse_game_json_dict(game_data: ZombieDieGameDict) -> ZombieDieGame:
    parameters = {
        key: value for key, value in game_data.items() if key in UNTRANSFORMED_KEYS
    }
    # ones we just need a simple model_validate_on
    parameters["commands"] = [
        (parse_command_dict(command), RoundState.model_validate(state))
        for command, state in game_data["commands"]
    ]
    parameters["players"] = [
        Player.model_validate(player) for player in game_data["players"]
    ]
    parameters["round"] = RoundState.model_validate(game_data.get("round"))

    # special fields to set.
    bag_func = game_data["bag_function"]
    if isinstance(bag_func, str) and bag_func != KnownFunctions.STANDARD:
        # We only allow us to use standard_bag when we load a dict.
        raise ValueError(
            f"Unable to understand the bag_function referenced: {bag_func}"
        )
    # default function for a game is standard anyway so this will just work lol.
    parameters["bag_function"] = None
    if not isinstance(bag_func, str):
        # basically we know these are recipes so load them.
        parameters["bag_recipes"] = [
            DieRecipe.model_validate(raw_recipe) for raw_recipe in bag_func
        ]

    return ZombieDieGame(**parameters)



"""
Module containing the code to run the cli for zombie dice and expose the first implemenation of the game.

If you want to just play a game using click's interface for getting users inputs from the cli 
you may use the `run_game` function.

```python
from zombie_nomnom import ZombieDieGame
from zombie_nomnom.cli import run_game


# Starts a full game from setup to full play by play.
run_game() 

existing_game = ZombieDieGame(players=["Me", "You", "Mr. McGee"])

# Run the cli for an already running/existing game.
run_game(existing_game)

```

If you would like to just use some of the functions we have to render different objects
you can use our print functions such as: `render_players`, `render_turn`, `render_winner`

```python
from zombie_nomnom import ZombieDieGame
from zombie_nomnom.cli import render_players, render_turn, render_winner

existing_game = ZombieDieGame(players=["Billy", "Zabka"])

# Prints out the details of the players of the game and their scores.
render_players(existing_game)

# Prints out the given round information object.
render_turn(existing_game.round)

# Prints out the highest scoring player 
# defaults to the player that went first in the case of a tie.
render_winner(existing_game)

```

This module is primarly used by app and should not be used by other parts of our library.

"""

from typing import Any, Callable, TypeVar
import click

from .engine import DrawDice, Player, RoundState, Score, ZombieDieGame

draw_command = DrawDice()
"""Command used to draw the dice required for a turn."""

score_command = Score()
"""Command used to score a players hand during the game."""


def draw_dice(game: ZombieDieGame):
    """Applys the DrawDice command to the game instance and
    renders the result to the console.

    **Parameters**
    - game (`zombie_nomnom.ZombieDieGame`): instance of the game to apply the draw command too.
    """
    click.echo("Drawing dice...")
    turn = game.process_command(draw_command)
    if not turn.ended:
        click.echo(_format_turn_info(turn))
    else:
        click.echo(f"Ohh no!! {turn.player.name} Has Died(T_T) R.I.P")


def score_hand(game: ZombieDieGame):
    """Applys the score command to a game instance and
    prints out result to the console.

    **Parameters**
    - game (`zombie_nomnom.ZombieDieGame`): game instance you want to score with.
    """
    click.echo("Scoring hand...")
    turn = game.process_command(score_command)
    click.echo(_format_turn_info(turn))


def exit(game: ZombieDieGame):
    """
    Exits the game by marking the game as over for players
    who wish to finish the current game they are playing
    and prints out to the console.

    **Parameters**
    - game (`zombie_nomnom.ZombieDieGame`): game instance you want to end.
    """
    click.echo("Ending game...")
    game.game_over = True


_actions: dict[str, Callable[[ZombieDieGame], None]] = {
    "Exit": exit,
    "Draw dice": draw_dice,
    "Score hand": score_hand,
}


def run_game(game: ZombieDieGame | None = None):
    """
    Main entrypoint for prompting and running the game,
    either an existing instance or creates a new one if not given.
    Will allow users to get prompted and play the game as well as run setup if no game
    is given.

    **Parameters**
    - game (`zombie_nomnom.ZombieDieGame` | `None`, optional): game instance that we want to run. Defaults to None.

    **Returns**
    - `zombie_nomnom.ZombieDieGame`: The instance of the game that has been played.
    """
    game = game or setup_game()
    while not game.game_over:
        # prime game with initial turn.
        render_players(game)
        play_turn(game)
    render_winner(game)
    return game


def render_winner(game: ZombieDieGame):
    """Prints out the current winner of the game instance to the console.

    **Parmeters**
    - game (`zombie_nomnom.ZombieDieGame`): instance that we are looking for the winner on.
    """
    formatted_player = _format_player(game.winner)
    click.echo(f"{formatted_player} Has Won!!")


def play_turn(game: ZombieDieGame):
    """Prompts the user on the console to select action for the turn and prints out the current turn information.

    **Parameters**
    - game (`zombie_nomnom.ZombieDieGame`): game we want to do a turn action on.
    """
    render_turn(game.round)
    select_dict_item(_actions)(game)


def _format_turn_info(turn: RoundState):
    player = turn.player
    bag = turn.bag

    return f"{player.name}, Hand: Brains({len(player.brains)}), Feet({len(player.rerolls)}), Shots({len(player.shots)}), Dice Remaining: {len(bag)}"


def render_turn(turn: RoundState):
    """Prints turn info to the console for a given RoundState.

    **Parameters**
    - turn (`zombie_nomnom.RoundState`): RoundState we are printing.
    """
    click.echo(f"Currently Playing {_format_turn_info(turn)}")


def _format_player(player: Player):
    """
    Formats a player object into a string for our rendering fuctions.

    **Parameters**
    - player (`zombie_nomnom.Player`): player we want to format as string

    **Returns**
    - `str`: Stringified version of player
    """
    return f"{player.name} ({player.total_brains})"


def render_players(game: ZombieDieGame):
    """Prints out the players currently playing in
    game instance. Will put them all on a single line.

    **Parameters**
    - game (`zombie_nomnom.ZombieDieGame`): game instance that we are getting players from.
    """
    players_listed = ", ".join(_format_player(player) for player in game.players)
    click.echo(f"Players: {players_listed}")


class StrippedStr(click.ParamType):
    """Custom `str` parameters that will take in the input from the cli
    and trim any trailing or leading spaces so that I can focus on just the value itself.
    """

    def convert(
        self, value: Any, param: click.Parameter | None, ctx: click.Context | None
    ) -> str:
        """Converts a value from clicks input function into a stripped `str`.
        If given an object will turn the object to a `str` using the `__str__` and then trim the output.

        **Parameters**
        - value (`Any`): Value that was taken by clicks input functions.
        - param (`click.Parameter` | `None`): Optional Parameter form click.
        - ctx (`click.Context` | `None`): Optional context form click.

        **Returns**
        - `str`: str value that has been trimmed.
        """
        if isinstance(value, str):
            return value.strip()
        else:
            return str(value).strip()


def setup_game() -> ZombieDieGame:
    """Runs the setup game cli prompts for users to enter the players in the game.
    This will prompt you for each player in the game then create and return the
    game instance you have with those players.

    **Returns**
    - `zombie_nomnom.ZombieDieGame`: The instance of the game you setup.
    """
    names = prompt_list(
        "Enter Player Name",
        _type=StrippedStr(),
        confirmation_prompt="Add Another Player?",
    )
    # TODO(Milo): Figure out a bunch of game types to play that we can use as templates for the die.
    return ZombieDieGame(
        players=[Player(name=name) for name in names],
    )


TVar = TypeVar("TVar")


def select_dict_item(value: dict[str, TVar]) -> TVar:
    """Prompts the user to select an item from a dictionary and
    then return the value stored at that key. The selection is
    based on an array of options that is orded the same as the
    way they are stored in the keys.

    **Parameters**
    - value (`dict[str, TVar]`): dictionary with values the user will select from.

    **Returns**
    - `TVar`: The value that was selected by the user.
    """
    menu_items = list(value)
    menu = "\n".join(f"{index}) {item}" for index, item in enumerate(menu_items))
    click.echo(menu)
    selected_index = click.prompt(
        f"Select Item (0-{len(menu_items) - 1})",
        type=click.IntRange(0, len(menu_items) - 1),
    )
    return value[menu_items[selected_index]]


def prompt_list(
    prompt: str,
    _type: type,
    confirmation_prompt: str = "Add Another?",
) -> list:
    """Prompts the user to input a list of items using the click library.
    Allows you to define the type of information you want in your list and then return that to you.

    **Parameters**
    - prompt (`str`): Prompt to display to the user.
    - _type (`type`): The type that you wish to get back, can also be a `click.ParamType`
    - confirmation_prompt (`str`, optional): Optional prompt for when the user is asked to add another item. Defaults to "Add Another?".

    **Returns**
    - `list`: Collection of items that the user has given.
    """
    inputs = []
    inputs.append(click.prompt(prompt, type=_type))

    while click.confirm(confirmation_prompt):
        inputs.append(click.prompt(prompt, type=_type))
    return inputs



"""
This module contains everything we care about when it comes to the die.
We define a class to contain the faces as well as be able to select one
using the random built-in library.

We can instatation an instance like this
```python
from zombie_nomnom.models.dice import Die, Face

custom_die = Die(faces=[
    Face.BRAIN,
    Face.BRAIN,
    Face.BRAIN,
    Face.SHOTGUN
    Face.SHOTGUN,
    Face.SHOTGUN,
])

# is currently set to None because it was not set.
custom_die.current_face 

# roll and calculate the current face
custom_die.roll()

# will be one of the faces in the in the faces array
custom_die.current_face
```

Most of the time you will want to use a preconfigured recipe
To build your dice which you can use our `create_die` method.

```python
from zombie_nomnom.models.dice import create_die, DieColor

green_die = create_die(DieColor.GREEN)

yellow_die = create_die(DieColor.YELLOW)

red_die = create_die(DieColor.RED)
```

Most of the die in the game are already pre-defined so you can expirement
and use different dice as you make your own custom games.

"""

from enum import Enum
from pydantic import BaseModel, Field, field_validator
import random


class Face(str, Enum):
    """
    Face of the die for the game.

    There are three core ones:
    - BRAIN: Single point scoring face
    - FOOT: Neutral dice that will be re-rolled first before any other die are given.
    - SHOTGUN: Damaging dice where you are limited to only so many before your turn is over.
    """

    BRAIN = "BRAIN"
    """Scoring `Face` worth a single point.
    """
    FOOT = "FOOT"
    """Neutral `Face` that we will reroll.
    """
    SHOTGUN = "SHOTGUN"
    """Damaging `Face` that may end the turn.
    """


class DieFace(BaseModel):
    """
    Represents a custom face on a die. Used to allow us to be able to score extra points or damage on a player when drawn.
    """

    name: str
    """Name of the face."""
    score: int
    """Amount of points a player will gain from face."""
    damage: int
    """Amount of damage a player will take from face."""


class DieColor(str, Enum):
    """
    Names of special core dice in the game.
    There are only three to begin with: RED, YELLOW, GREEN
    """

    RED = "RED"
    """The hardest die to score on in the game with only a single side.
    """
    YELLOW = "YELLOW"
    """The most medium die to score on with an equal number of brain and shot sides.
    """
    GREEN = "GREEN"
    """The most forgiving dice with only a single side that will damage you.
    """


# TODO(Milo): Update this for custom exception on invalid dice.
class Die(BaseModel):
    """
    Represents the die we are rolling in the game.
    This is currently enforced to only support 6 sided dice.
    """

    name: str | None = None
    """
        The plaintext name of the die.
    """
    faces: list[DieFace | Face] = Field(min_length=6, max_length=6)
    """
    faces of the dice. It is currently only allowed to have 6 values.
    """
    current_face: Face | DieFace | None = None
    """
    The currently displayed face of the die. Defaults to None.
    """

    def roll(self) -> Face:
        """Rolls the dice using the `builtins.random` and updates the current_face field.

        **Returns**
        - `Face`: The face that the die is now on.
        """
        self.current_face = random.choice(self.faces)
        return self.current_face


_dice_face_mapping = {
    DieColor.RED: {Face.BRAIN: 1, Face.FOOT: 2, Face.SHOTGUN: 3},
    DieColor.YELLOW: {Face.BRAIN: 2, Face.FOOT: 2, Face.SHOTGUN: 2},
    DieColor.GREEN: {Face.BRAIN: 3, Face.FOOT: 2, Face.SHOTGUN: 1},
}


# TODO(Milo): Update this with custom exception for invalid dice.
def create_die(color: DieColor) -> Die:
    """Factory method to take in a `DieColor` and then create a die based on that.
    Only supports the first three colors defined: `DieColor.RED`, `DieColor.YELLOW`, `DieColor.GREEN`.

    **Parameters**
    - color (`DieColor`): The color of the dice you want the factory to produce.

    `Raises`:
    - `ValueError`: When we are unable to resolve the color to a known recipe.

    `Returns`
    - `Die`: The dice defined by the color given.
    """
    if color not in _dice_face_mapping:
        raise ValueError(f"Unknown Die Color: {color}")

    mapped_color = _dice_face_mapping[color]
    faces = []
    for face, amount in mapped_color.items():
        faces.extend([face] * amount)
    return Die(faces=faces, name=color)



from zombie_nomnom.engine import DrawDice, Score, ZombieDieGame
from zombie_nomnom.models.dice import Face


def test_zombie_dice_game__plays_a_valid_single_player_game_of_zombie_dice__finishes_in_my_victory(
    patch_random_randint,
    patch_random_choice,
):
    # basically always pick the first three die in the array and then default to always the first face on each die.
    patch_random_randint.side_effect = [
        *[0, 1, 2],
        *[0, 1, 2],
        *[0, 1, 2],
        *[0, 1, 2],
        *[0, 1, 2],
    ]
    patch_random_choice.side_effect = lambda collection: collection[0]

    draw_three = DrawDice()
    score = Score()

    game = ZombieDieGame(
        players=["Game Ova"],
    )

    game.process_command(draw_three)
    game.process_command(draw_three)
    game.process_command(draw_three)
    game.process_command(draw_three)
    game.process_command(draw_three)
    game.process_command(score)

    assert game.winner.total_brains == 15
    assert game.game_over, "Game should be finished"
    assert game.winner.name == "Game Ova"


def test_zombie_dice_game__plays_a_valid_two_player_game_of_zombie_dice__finishes_in_player_one_victory(
    patch_random_randint,
    patch_random_choice,
):
    # basically always pick the first three die in the array and then default to always the first face on each die.
    patch_random_randint.side_effect = [
        *[0, 1, 2],
        *[0, 1, 2],
        *[0, 1, 2],
        *[0, 1, 2],
        *[0, 1, 2],
        *[0, 1, 2],
        *[0, 1, 2],
        *[0, 1, 2],
        *[0, 1, 2],
        *[0, 1, 2],
    ]
    patch_random_choice.side_effect = [
        *([Face.BRAIN] * 15),
        *([Face.BRAIN] * 14),
        Face.SHOTGUN,
    ]

    draw_three = DrawDice()
    score = Score()

    game = ZombieDieGame(
        players=[
            "Game Ova",
            "Looser",
        ],
    )

    # First Players Turn
    game.process_command(draw_three)
    game.process_command(draw_three)
    game.process_command(draw_three)
    game.process_command(draw_three)
    game.process_command(draw_three)
    game.process_command(score)

    # Second Players Turn
    game.process_command(draw_three)
    game.process_command(draw_three)
    game.process_command(draw_three)
    game.process_command(draw_three)
    game.process_command(draw_three)
    game.process_command(score)

    assert game.winner.total_brains == 15
    assert game.game_over, "Game should be finished"
    assert game.winner.name == "Game Ova"



from typing import Callable
from click.testing import CliRunner, Result
import pytest

from zombie_nomnom.app import main


@pytest.fixture
def runner():
    return CliRunner()


@pytest.fixture
def run_game_with_input(runner: CliRunner):
    def _run_with_input(cli_input: str = "") -> Result:
        return runner.invoke(main, args=["cli"], input=cli_input)

    return _run_with_input


def test_app__when_user_types_name__prints_name_to_screen(
    run_game_with_input: Callable[[str], Result]
):
    # this works by putting in the name of the player then new line to tell it to no more players then zero
    # to select exit option and then final new line to not continue
    cli_input = """milo

    0
    """
    result = run_game_with_input(cli_input)
    assert "Players: milo" in result.output


def test_app__when_replaying_game__setups_game_again(
    run_game_with_input: Callable[[str], Result]
):
    # this works by putting in the name of the player then new line to tell it to no more players then zero
    # to select exit option and then final new line to not continue
    cli_input = """milo

    0
    y
    milo
    
    0
    n
    """
    result = run_game_with_input(cli_input)
    assert "Players: milo" in result.output


def test_app__when_setting_up__allows_multiple_players(
    run_game_with_input: Callable[[str], Result]
):
    # this works by putting in the name of the player then new line to tell it to no more players then zero
    # to select exit option and then final new line to not continue
    cli_input = """milo
y
Dean

    0
    """
    result = run_game_with_input(cli_input)
    assert "Players: milo" in result.output
    assert "Dean (0)" in result.output


def test_app__when_playing_and_drawing_dice__dice_goes_down_by_3_for_first_draw(
    run_game_with_input: Callable[[str], Result]
):
    # this works by putting in the name of the player then new line to tell it to no more players then zero
    # to select exit option and then final new line to not continue
    cli_input = """milo
    
    1
    0
    """
    result = run_game_with_input(cli_input)
    assert "Drawing dice..." in result.output, "Did not display drawing dice"
    assert (
        "Dice Remaining: 10" in result.output
    ), "Did not take three dice correctly from the game."


def test_app__when_playing_and_drawing_dice__dice_goes_down_by_3_for_first_draw(
    run_game_with_input: Callable[[str], Result]
):
    # this works by putting in the name of the player then new line to tell it to no more players then zero
    # to select exit option and then final new line to not continue
    cli_input = """milo
    
    1
    0
    """
    result = run_game_with_input(cli_input)
    assert "Drawing dice..." in result.output, "Did not display drawing dice"
    assert (
        "Dice Remaining: 10" in result.output
    ), "Did not take three dice correctly from the game."


def test_app__when_playing_and_scoring_hand__scores_dice_and_transitions_turn(
    run_game_with_input: Callable[[str], Result]
):
    # this works by putting in the name of the player then new line to tell it to no more players then zero
    # to select exit option and then final new line to not continue
    cli_input = """milo
    y
    dean
    n
    2
    0
    """
    result = run_game_with_input(cli_input)
    assert "Scoring hand..." in result.output, "Did not display scoring hand..."
    assert "Currently Playing dean" in result.output, "Did not display deans turn..."


def test_app__when_playing_and_scoring_hand__scores_dice_and_transitions_turn(
    run_game_with_input: Callable[[str], Result],
):
    # this works by putting in the name of the player then new line to tell it to no more players then zero
    # to select exit option and then final new line to not continue
    cli_input = """milo
    y
    dean
    n
    2
    0
    """
    result = run_game_with_input(cli_input)
    assert "Scoring hand..." in result.output, "Did not display scoring hand..."
    assert "Currently Playing dean" in result.output, "Did not display deans turn..."


def test_app__when_playing_and_rolling_until_death__transitions_turn_to_other_player(
    run_game_with_input: Callable[[str], Result]
):
    # this works by putting in the name of the player then new line to tell it to no more players then zero
    # to select exit option and then final new line to not continue
    cli_input = """milo
    y
    dean

    1
    1
    1
    1
    1
    1
    1
    1
    1
    0
    """
    result = run_game_with_input(cli_input)
    assert "milo Has Died" in result.output, "He is a god gamer and survived."
